# Note: The classes in this module are not to be re-exported by `__init__.py`,
# as that would mean unconditionally importing setuptools whenever versioningit
# is imported, slowing things down.

from .get_cmdclasses import get_cmdclasses

_cmdclasses = get_cmdclasses()
sdist = _cmdclasses["sdist"]
build_py = _cmdclasses["build_py"]



from __future__ import annotations
from dataclasses import fields
from pathlib import Path
import shutil
import tempfile
from typing import Any
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl
from hatchling.version.source.plugin.interface import VersionSourceInterface
from .config import Config
from .core import Report, Versioningit
from .errors import NoTagError, NotSdistError, NotVersioningitError
from .logging import init_logging, log
from .onbuild import HatchFileProvider


class VersioningitSource(VersionSourceInterface):
    PLUGIN_NAME = "versioningit"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__fields: dict[str, Any] | None = None

    def get_version_data(self) -> dict:
        """
        The entry point called by hatch to retrieve the version for a project
        """
        init_logging()
        PROJECT_ROOT = Path(self.root)
        log.info("Project dir: %s", PROJECT_ROOT)
        try:
            vgit = Versioningit.from_project_dir(PROJECT_ROOT)
            report = vgit.run(write=True, fallback=True)
        except NotVersioningitError:  # pragma: no cover
            p = PROJECT_ROOT / "pyproject.toml"
            raise RuntimeError(f"versioningit not configured in {p}")
        except (NotSdistError, NoTagError) as e:
            raise RuntimeError(
                # If an error occurs in `get_version_data()`, hatchling throws
                # away its cause, so we need to include `str(e)` in the
                # RuntimeError for it to be seen.
                f"{type(e).__name__}: {e}\n"
                "\nversioningit could not find a version for the project in"
                f" {PROJECT_ROOT}!\n\n"
                "You may be installing from a shallow clone, in which case you"
                " need to unshallow it first.\n\n"
                "Alternatively, you may be installing from a Git archive, which is"
                " not supported by default.  Install from a git+https://... URL"
                " instead.\n\n"
            )
        if isinstance(report, Report):
            self.__fields = report.template_fields
        return {"version": report.version}

    def set_version(
        self, _version: str, _version_data: dict
    ) -> None:  # pragma: no cover
        # Can't be tested due to hatch using isolated environments for `hatch
        # version`
        raise NotImplementedError(
            "The versioningit plugin does not support setting the version."
            "  Create a tag instead."
        )

    def get_template_fields(self) -> dict[str, Any] | None:
        return self.__fields


class OnbuildHook(BuildHookInterface):
    PLUGIN_NAME = "versioningit-onbuild"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__tmpdir = Path(tempfile.mkdtemp())

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        init_logging()
        if self.target_name == "wheel" and version == "editable":
            log.debug("Not running onbuild step for editable build")
            return None
        version_source = self.metadata.hatch.version.source
        if not isinstance(version_source, VersioningitSource):
            raise RuntimeError(
                "versioningit-onbuild can only be used with 'versioningit'"
                " version source, but version source is"
                f" {self.metadata.hatch.version.source_name!r}"
            )
        template_fields = version_source.get_template_fields()
        if template_fields is None:
            log.debug("Appear to be building from an sdist; not running onbuild step")
            return
        config = self.config.copy()
        for key in [
            "dependencies",
            "require-runtime-dependencies",
            "require-runtime-features",
        ]:
            config.pop(key, None)
        (onbuild_field,) = [f for f in fields(Config) if f.name == "onbuild"]
        onbuild_section = Config.parse_section(onbuild_field, config)
        assert onbuild_section is not None
        onbuild_method = onbuild_section.load(self.root)
        file_provider = HatchFileProvider(
            src_dir=Path(self.root),
            tmp_dir=self.__tmpdir,
        )
        onbuild_method(
            file_provider=file_provider,
            is_source=self.target_name == "sdist",
            template_fields=template_fields,
        )
        build_data.setdefault("force_include", {}).update(
            file_provider.get_force_include()
        )

    def finalize(
        self, _version: str, _build_data: dict[str, Any], _artifact_path: str
    ) -> None:
        shutil.rmtree(self.__tmpdir, ignore_errors=True)


@hookimpl
def hatch_register_version_source() -> type[VersionSourceInterface]:
    # This function must be named "hatch_register_version_source" exactly.
    return VersioningitSource


@hookimpl
def hatch_register_build_hook() -> type[BuildHookInterface]:
    return OnbuildHook



"""
Versioning It with your Version In Git

``versioningit`` is yet another Python packaging plugin for automatically
determining your package's version based on your version control repository's
tags.  Unlike others, it allows easy customization of the version format and
even lets you easily override the separate functions used for version
extraction & calculation.

**Features:**

- Works with both setuptools and Hatch_

  .. _hatch: https://hatch.pypa.io

- Installed & configured through :pep:`518`'s ``pyproject.toml``

- Supports Git, modern Git archives, and Mercurial

- Formatting of the final version uses format template strings, with fields for
  basic VCS information and separate template strings for distanced vs. dirty
  vs. distanced-and-dirty repository states

- Can optionally write the final version to a file for loading at runtime

- Provides custom hooks for inserting the final version into a source file at
  build time

- The individual methods for VCS querying, tag-to-version calculation, version
  bumping, version formatting, and writing the version to a file can all be
  customized using either functions defined alongside one's project code or via
  publicly-distributed entry points

- Can alternatively be used as a library for use in ``setup.py`` or the like,
  in case you don't want to or can't configure it via ``pyproject.toml``

- The only thing it does is calculate your version and optionally write it to a
  file; there's no overriding of your sdist contents based on what's in your
  Git repository, especially not without a way to turn it off, because that
  would just be rude.

Visit <https://github.com/jwodder/versioningit> or
<https://versioningit.rtfd.io> for more information.
"""

__version__ = "3.1.2"
__author__ = "John Thorvald Wodder II"
__author_email__ = "versioningit@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/versioningit"

from .core import (
    FallbackReport,
    Report,
    VCSDescription,
    Versioningit,
    get_next_version,
    get_template_fields_from_distribution,
    get_version,
    get_version_from_pkg_info,
    run_onbuild,
)
from .errors import (
    ConfigError,
    Error,
    InvalidTagError,
    InvalidVersionError,
    MethodError,
    NoTagError,
    NotSdistError,
    NotVCSError,
    NotVersioningitError,
)
from .get_cmdclasses import get_cmdclasses
from .onbuild import OnbuildFile, OnbuildFileProvider

__all__ = [
    "ConfigError",
    "Error",
    "FallbackReport",
    "InvalidTagError",
    "InvalidVersionError",
    "MethodError",
    "NoTagError",
    "NotSdistError",
    "NotVCSError",
    "NotVersioningitError",
    "OnbuildFile",
    "OnbuildFileProvider",
    "Report",
    "VCSDescription",
    "Versioningit",
    "get_cmdclasses",
    "get_next_version",
    "get_template_fields_from_distribution",
    "get_version",
    "get_version_from_pkg_info",
    "run_onbuild",
]



from __future__ import annotations
from abc import ABC, abstractmethod
from contextlib import suppress
from dataclasses import dataclass, field
import os
from pathlib import Path, PurePath
import re
import shutil
import tempfile
from typing import IO, TYPE_CHECKING, Any, TextIO, overload
from .errors import ConfigError
from .logging import log, warn_extra_fields
from .util import bool_guard, ensure_terminated, optional_str_guard, str_guard

if TYPE_CHECKING:
    from typing_extensions import Literal, TypeAlias

    TextMode: TypeAlias = Literal["r", "w", "a"]
    BinaryMode: TypeAlias = Literal["rb", "br", "wb", "bw", "ab", "ba"]


class OnbuildFileProvider(ABC):
    """
    .. versionadded:: 3.0.0

    An abstract base class for accessing files that are about to be included in
    an sdist or wheel currently being built
    """

    @abstractmethod
    def get_file(
        self, source_path: str | PurePath, install_path: str | PurePath, is_source: bool
    ) -> OnbuildFile:
        """
        Get an object for reading & writing a file in the project being built.

        :param source_path:
            the path to the file relative to the root of the project's source
        :param install_path:
            the path to the same file when it's in a wheel, relative to the
            root of the wheel (or, equivalently, the path to the file when it's
            installed in a site-packages directory, relative to that directory)
        :param is_source:
            `True` if building an sdist or other artifact that preserves source
            paths, `False` if building a wheel or other artifact that uses
            install paths
        """
        ...


class OnbuildFile(ABC):
    """
    .. versionadded:: 3.0.0

    An abstract base class for opening a file in a project currently being
    built
    """

    @overload
    def open(
        self,
        mode: TextMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        mode: BinaryMode,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> IO[bytes]: ...

    @abstractmethod
    def open(
        self,
        mode: TextMode | BinaryMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """
        Open the associated file.  ``mode`` must be ``"r"``, ``"w"``, ``"a"``,
        ``"rb"``, ``"br"``, ``"wb"``, ``"bw"``, ``"ab"``, or ``"ba"``.

        When opening a file for writing or appending, if the file does not
        already exist, any parent directories are created automatically.
        """
        ...


@dataclass
class SetuptoolsFileProvider(OnbuildFileProvider):
    """
    .. versionadded:: 3.0.0

    `OnbuildFileProvider` implementation for use when building sdists or wheels
    under setuptools.

    Setuptools builds its artifacts by creating a temporary directory
    containing all of the files (sometimes hardlinked) that will go into them
    and then building an archive from that directory.  "onbuild" runs just
    before the archive step, so this provider simply operates directly on the
    temporary directory without ever looking at the project source.
    """

    #: The setuptools-managed temporary directory containing the files for the
    #: archive currently being built
    build_dir: Path

    #: The set of file paths in `build_dir` (relative to `build_dir`) that have
    #: been opened for writing or appending
    modified: set[PurePath] = field(init=False, default_factory=set)

    def get_file(
        self, source_path: str | PurePath, install_path: str | PurePath, is_source: bool
    ) -> SetuptoolsOnbuildFile:
        return SetuptoolsOnbuildFile(
            provider=self,
            source_path=PurePath(source_path),
            install_path=PurePath(install_path),
            is_source=is_source,
        )


@dataclass
class SetuptoolsOnbuildFile(OnbuildFile):
    provider: SetuptoolsFileProvider
    source_path: PurePath
    install_path: PurePath
    is_source: bool

    @overload
    def open(
        self,
        mode: TextMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        mode: BinaryMode,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> IO[bytes]: ...

    def open(
        self,
        mode: TextMode | BinaryMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        path = self.source_path if self.is_source else self.install_path
        p = self.provider.build_dir / path
        if ("w" in mode or "a" in mode) and path not in self.provider.modified:
            self.provider.modified.add(path)
            p.parent.mkdir(parents=True, exist_ok=True)
            # If setuptools is using hard links for the build files, undo that
            # for this file:
            if "w" in mode:
                with suppress(FileNotFoundError):
                    p.unlink()
            elif p.exists():
                # We've been asked to append to the file, so replace it with a
                # non-hardlinked copy of its contents:
                fd, tmp = tempfile.mkstemp(dir=self.provider.build_dir)
                os.close(fd)
                shutil.copy2(p, tmp)
                os.replace(tmp, p)
        return p.open(mode=mode, encoding=encoding, errors=errors, newline=newline)


@dataclass
class HatchFileProvider(OnbuildFileProvider):
    """
    .. versionadded:: 3.0.0

    `OnbuildFileProvider` implementation for use when building sdists or wheels
    under Hatch.

    Hatch builds its artifacts by reading the contents of the files in the
    project directory directly into an in-memory archive.  In order to modify
    what goes into that archive without altering anything in the project
    directory, we need to write all modifications to a temporary directory and
    register the resulting files as "forced inclusion paths."
    """

    #: The root of the project directory
    src_dir: Path

    #: A temporary directory (managed outside the provider) in which to create
    #: modified files
    tmp_dir: Path

    #: The set of file paths created under the temporary directory, relative to
    #: the temporary directory
    modified: set[PurePath] = field(init=False, default_factory=set)

    def get_file(
        self, source_path: str | PurePath, install_path: str | PurePath, is_source: bool
    ) -> HatchOnbuildFile:
        return HatchOnbuildFile(
            provider=self,
            source_path=PurePath(source_path),
            install_path=PurePath(install_path),
            is_source=is_source,
        )

    def get_force_include(self) -> dict[str, str]:
        return {str(self.tmp_dir / p): str(p) for p in self.modified}


@dataclass
class HatchOnbuildFile(OnbuildFile):
    provider: HatchFileProvider
    source_path: PurePath
    install_path: PurePath
    is_source: bool

    @overload
    def open(
        self,
        mode: TextMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        mode: BinaryMode,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> IO[bytes]: ...

    def open(
        self,
        mode: TextMode | BinaryMode = "r",
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        path = self.source_path if self.is_source else self.install_path
        if "r" in mode and path not in self.provider.modified:
            return (self.provider.src_dir / self.source_path).open(
                mode=mode, encoding=encoding, errors=errors
            )
        else:
            p = self.provider.tmp_dir / path
            if ("w" in mode or "a" in mode) and path not in self.provider.modified:
                self.provider.modified.add(path)
                p.parent.mkdir(parents=True, exist_ok=True)
                if not p.exists() and "a" in mode:
                    with suppress(FileNotFoundError):
                        shutil.copy2(self.provider.src_dir / self.source_path, p)
            return p.open(mode=mode, encoding=encoding, errors=errors, newline=newline)


def replace_version_onbuild(
    *,
    file_provider: OnbuildFileProvider,
    is_source: bool,
    template_fields: dict[str, Any],
    params: dict[str, Any],
) -> None:
    """Implements the ``"replace-version"`` ``onbuild`` method"""

    DEFAULT_REGEX = r"^\s*__version__\s*=\s*(?P<version>.*)"
    DEFAULT_REPLACEMENT = '"{version}"'

    params = params.copy()
    source_file = str_guard(params.pop("source-file", None), "onbuild.source-file")
    build_file = str_guard(params.pop("build-file", None), "onbuild.build-file")
    encoding = str_guard(params.pop("encoding", "utf-8"), "onbuild.encoding")
    regex = str_guard(params.pop("regex", DEFAULT_REGEX), "onbuild.regex")
    try:
        rgx = re.compile(regex)
    except re.error as e:
        raise ConfigError(f"versioningit: onbuild.regex: Invalid regex: {e}")
    require_match = bool_guard(
        params.pop("require-match", False), "onbuild.require-match"
    )
    replacement = str_guard(
        params.pop("replacement", DEFAULT_REPLACEMENT), "onbuild.replacement"
    )
    append_line = optional_str_guard(
        params.pop("append-line", None), "onbuild.append-line"
    )
    warn_extra_fields(
        params,
        "onbuild",
        [
            "source-file",
            "build-file",
            "encoding",
            "regex",
            "require-match",
            "replacement",
            "append-line",
        ],
    )

    path = source_file if is_source else build_file
    log.info("Updating version in file %s", path)
    file = file_provider.get_file(
        source_path=source_file,
        install_path=build_file,
        is_source=is_source,
    )
    with file.open(encoding=encoding) as fp:
        # Don't use readlines(), as that doesn't split on everything that
        # splitlines() uses
        lines = fp.read().splitlines(keepends=True)
    for i, ln in enumerate(lines):
        m = rgx.search(ln)
        if m:
            log.debug("onbuild.regex matched file on line %d", i + 1)
            vgroup: str | int
            if "version" in m.groupdict():
                vgroup = "version"
            else:
                vgroup = 0
            if m[vgroup] is None:
                raise RuntimeError(
                    "'version' group in versioningit's onbuild.regex did"
                    " not participate in match"
                )
            newline = ensure_terminated(
                ln[: m.start(vgroup)]
                + m.expand(replacement.format_map(template_fields))
                + ln[m.end(vgroup) :]
            )
            log.debug("Replacing line %r with %r", ln, newline)
            lines[i] = newline
            break
    else:
        if require_match:
            raise RuntimeError(f"onbuild.regex did not match any lines in {path}")
        elif append_line is not None:
            log.info(
                "onbuild.regex did not match any lines in the file; appending line"
            )
            if lines:
                lines[-1] = ensure_terminated(lines[-1])
            lines.append(ensure_terminated(append_line.format_map(template_fields)))
        else:
            log.info(
                "onbuild.regex did not match any lines in the file; leaving unmodified"
            )
            return
    with file.open("w", encoding=encoding) as fp:
        fp.writelines(lines)



from __future__ import annotations
from dataclasses import Field, dataclass, field, fields
from pathlib import Path
import sys
from typing import Any, Optional
from .errors import ConfigError, NotVersioningitError
from .logging import log, warn_extra_fields
from .methods import (
    CallableSpec,
    CustomMethodSpec,
    EntryPointSpec,
    MethodSpec,
    VersioningitMethod,
)
from .util import optional_str_guard

if sys.version_info[:2] >= (3, 11):
    from tomllib import load as toml_load
else:
    from tomli import load as toml_load


@dataclass
class ConfigSection:
    """A parsed method subtable of the `versioningit` configuration"""

    #: Specification of the method for this subtable
    method_spec: MethodSpec

    #: Additional parameters to pass to the method
    params: dict[str, Any]

    def load(self, project_dir: str | Path) -> VersioningitMethod:
        """Loads the method and returns a `VersioningitMethod`"""
        return VersioningitMethod(self.method_spec.load(project_dir), self.params)


@dataclass
class Config:
    """Parsed `versioningit` configuration"""

    #: Parsed ``vcs`` subtable
    vcs: ConfigSection = field(metadata={"default_entry_point": "git"})

    #: Parsed ``tag2version`` subtable
    tag2version: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``next-version`` subtable
    next_version: ConfigSection = field(metadata={"default_entry_point": "minor"})

    #: Parsed ``format`` subtable
    format: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``template-fields`` subtable
    template_fields: ConfigSection = field(metadata={"default_entry_point": "basic"})

    #: Parsed ``write`` subtable
    write: Optional[ConfigSection] = field(
        default=None, metadata={"default_entry_point": "basic", "optional": True}
    )

    #: Parsed ``onbuild`` subtable
    onbuild: Optional[ConfigSection] = field(
        default=None,
        metadata={"default_entry_point": "replace-version", "optional": True},
    )

    #: The ``default-version`` setting
    default_version: Optional[str] = None

    @classmethod
    def parse_toml_file(cls, filepath: str | Path) -> Config:
        """
        Parse the given TOML file and extract the contents of either the
        ``[tool.versioningit]`` table or the ``[tool.hatch.version]`` table (if
        it contains a ``source`` key set to ``"versioningit"``).

        If the ``[tool.versioningit]`` table is present and the
        ``[tool.hatch.version]`` table has more than just a ``source`` key,
        then a warning is emitted and the latter table is used.

        :raises NotVersioningitError:
            if the file does not contain a versioningit configuration table
        :raises ConfigError:
            if the configuration table or any of its subfields are not of the
            correct type
        """
        with open(filepath, "rb") as fp:
            tool = toml_load(fp).get("tool", {})
        table = tool.get("versioningit")
        try:
            hatch_config = tool["hatch"]["version"]
        except (AttributeError, LookupError, TypeError):
            pass
        else:
            if (
                isinstance(hatch_config, dict)
                and hatch_config.get("source") == "versioningit"
            ):
                hatch_config.pop("source", None)
                if hatch_config:
                    if table is not None:
                        log.warning(
                            "versioningit configuration found in both"
                            " [tool.hatch.version] and [tool.versioningit];"
                            " only using the former"
                        )
                    table = hatch_config
                elif table is None:
                    table = hatch_config
                if table.get("onbuild") is not None:
                    log.warning(
                        "onbuild configuration in versioningit table detected."
                        "  When using Hatch, onbuild must be configured via"
                        " [tool.hatch.build.hooks.versioningit-onbuild]."
                    )
        if table is None:
            raise NotVersioningitError("versioningit not enabled in pyproject.toml")
        return cls.parse_obj(table)

    @classmethod
    def parse_obj(cls, obj: Any) -> Config:
        """
        Parse a raw Python configuration structure

        :raises ConfigError:
            - if ``obj`` is not a `dict`
            - if ``default-version`` or any of the subtable or ``method``
              fields are not of the correct type
        """
        if not isinstance(obj, dict):
            raise ConfigError("versioningit config must be a table")
        default_version = optional_str_guard(
            obj.pop("default-version", None), "default-version"
        )
        sections: dict[str, Optional[ConfigSection]] = {}
        for f in fields(cls):
            if not f.metadata:
                continue
            sections[f.name] = cls.parse_section(f, obj.pop(attr2key(f.name), None))
        warn_extra_fields(
            obj,
            None,
            [attr2key(f.name) for f in fields(cls)],
        )
        return cls(
            default_version=default_version, **sections  # type: ignore[arg-type]
        )

    @staticmethod
    def parse_section(f: Field, obj: Any) -> Optional["ConfigSection"]:
        """
        Parse a step configuration field according to the metadata in the given
        `dataclasses.Field`, which must consist of the following items:

        ``default_entry_point`` : string
            The name of the default method to use for the step if one is not
            specified in the configuration

        ``optional`` : bool
            If true, an absent/`None` step field will be converted to `None`
            instead of to a section with the default method and no user
            parameters.  If not specified, defaults to false.

        :raises ConfigError:
            - if ``obj`` is not `None`, a callable, a string, or a `dict`
            - if any of the ``method`` fields are not of the correct type
        """
        if obj is None:
            if f.metadata.get("optional"):
                return None
            else:
                obj = f.metadata["default_entry_point"]
        if isinstance(obj, str):
            method_spec = Config.parse_method_spec(f, obj)
            return ConfigSection(method_spec, {})
        elif callable(obj):
            method_spec = CallableSpec(obj)
            return ConfigSection(method_spec, {})
        elif isinstance(obj, dict):
            if "method" not in obj and "module" in obj and "value" in obj:
                method_spec = Config.parse_method_spec(f, obj)
                return ConfigSection(method_spec, {})
            method_spec = Config.parse_method_spec(f, obj.pop("method", None))
            return ConfigSection(method_spec, obj)
        else:
            raise ConfigError(
                f"versioningit: {attr2key(f.name)} must be a string or table"
            )

    @staticmethod
    def parse_method_spec(f: Field, method: Any) -> MethodSpec:
        """
        Parse a ``method`` field according to the metadata in the given
        `dataclasses.Field` (see `parse_section()`)

        :raises ConfigError:
            - if ``method`` is not `None`, a string, a callable, or a `dict`
            - if ``method`` is a `dict` without a ``module`` or ``value`` key
            - if any of the fields in ``method`` are not of the correct type
        """
        key = attr2key(f.name)
        if method is None:
            method = f.metadata["default_entry_point"]
        if isinstance(method, str):
            return EntryPointSpec(group=f"versioningit.{f.name}", name=method)
        elif callable(method):
            return CallableSpec(method)
        elif isinstance(method, dict):
            module = method.pop("module", None)
            if not isinstance(module, str):
                raise ConfigError(
                    f"versioningit: {key}.method.module is required and"
                    " must be a string"
                )
            value = method.pop("value", None)
            if not isinstance(value, str):
                raise ConfigError(
                    f"versioningit: {key}.method.value is required and"
                    " must be a string"
                )
            module_dir = method.pop("module-dir", None)
            if module_dir is not None and not isinstance(module_dir, str):
                raise ConfigError(
                    f"versioningit: {key}.method.module-dir must be a string"
                )
            warn_extra_fields(
                method,
                f"{key}.method",
                ["module", "value", "module-dir"],
            )
            return CustomMethodSpec(module, value, module_dir)
        else:
            raise ConfigError(f"versioningit: {key}.method must be a string or table")


def attr2key(name: str) -> str:
    return name.replace("_", "-")
